/**
 * @file cpu.c
 * The core of processor, this file contains instruction decoder
 * and memory structure. Before use, call @a Reset() to set correct values
 * of registers. Execution of one instruction is done by calling @a InstrStep().
 *
 * Known bugs:
 * <ul>
 * <li />When executing instructions in the IO area, wrong instruction timing
 * can be computed (the IO routine might be called once for computing cycles
 * to be consumed and once when instructions arguments are obtained.) But
 * this bug shouldn't appear huh?
 * </ul>
 */

#include "cpu.h"


#define TRUE 1
#define FALSE 0

#define PAGESIZE 8

/** @name Flags */
/*@{*/
#define FN 0x80		/**< Negative flag */
#define FV 0x40		/**< Overflow flag */
#define FB 0x10		/**< BRK flag */
#define FD 0x08		/**< Decimal flag */
#define FI 0x04		/**< Interrupt flag */
#define FZ 0x02		/**< Zero flag */
#define FC 0x01		/**< Caryy flag */
/*@}*/


/** @name Adresses */
/*@{*/
#define RESETVKT 0xFFFC	/**< Reset interrupt jumps here */
#define IRQVKT 0xFFFE	/**< IRQ jumps here */
#define NMIVKT 0xFFFA
/*@}*/

/** @name Flag handling defines */
/*@{*/
/** Gets value of flag X, for conditions */
#define GETFLAG(X)	\
	((regs.P & (X))!=0)
/** Sets flag X to 1. Affects P register */
#define SETFLAG(X)	\
	regs.P=((regs.P) | (X))
/** Sets flag X to 0. Affects P register */
#define CLFLAG(X)	\
	(regs.P=regs.P & (~(X)))
/** If condition BL is satisfied, flag FL is set to 1 otherwise 0*/
#define FLAG(BL, FL)	\
	if (BL) SETFLAG(FL); else CLFLAG(FL)
/** Sets N and Z flag according to X */
#define SETNZFLAGS(X)	\
	FLAG((X) & 0x80, FN); FLAG(!(X), FZ)
/*@}*/

/** @name Other defines */
/*@{*/
/** Converts 2 bytes to 16-bit adress */
#define ADR(X, Y)	\
	(((Y) << PAGESIZE)+(X))
/** Sets PC to 2b adress stored in memory from adress X */
#define M2PC(X)	\
	regs.PC=((*memgetb)((X) + 1) << PAGESIZE) + (*memgetb)(X)
/** Higher byte of word */
#define HI(X)	\
	((unsigned char) (((X) >> PAGESIZE) & 0xFF))
/** Lower byte of word */
#define LO(X)	\
	((unsigned char) ((X) & 0xFF))
/** Moves execution point */
#define PCSTEP(X)	\
	regs.PC+=(X)
/** Converts number to BCD. Example: 10 dec -> $10 hex */
#define HEX2DEC(X)	\
	((((X)>>4) & 0x0F)*10 + ((X) & 0x0F))
/** Converts number in BCD code to std. decimal. Example> $10 hex -> 10 dec*/
#define DEC2HEX(X)	\
	(((((X)/10) % 10)<<4) + (X) % 10)
/*@}*/


/** @name Stack handling defines */
/*@{*/
/** Adds byte X to/from the top of stack*/
#define STACKADD(X)	\
	(*memsetb)(regs.S, (X)); regs.S--; if (regs.S<0x100) regs.S=0x1FF
/** Removes byte X to/from the top of stack*/
#define STACKREMOVE(X)	\
	regs.S++; if (regs.S>0x1FF) regs.S=0x100; X=(*memgetb)(regs.S)
/*@}*/

/** @name Cycle consumption defines.
 * Defines computing cycle consumption according to addresing method. Affects only several instructions
 */
/*@{*/
/** Consume X cycles */
#define CONSUME(X)	\
	cycles += X
/** If branch crosses a page boundary, add 1 cycle */
#define CONSREL	\
	if ((oldpc & 0xFF00) != (regs.PC & 0xFF00)) CONSUME(1)
/** If page boundary is crossed, add 1 cycle */
#define CONSABSX	\
	if ((ADR((*memgetb)(regs.PC+1), (*memgetb)(regs.PC+2)) & 0xFF00) !=	\
		((ADR((*memgetb)(regs.PC+1), (*memgetb)(regs.PC+2)) + regs.X) & 0xFF00)) \
		CONSUME(1)
/** If page boundary is crossed, add 1 cycle */
#define CONSABSY	\
	if ((ADR((*memgetb)(regs.PC+1), (*memgetb)(regs.PC+2)) & 0xFF00) !=	\
		((ADR((*memgetb)(regs.PC+1), (*memgetb)(regs.PC+2)) + regs.Y) & 0xFF00)) \
		CONSUME(1)
/** If page boundary is crossed, add 1 cycle */
#define CONSINDY	\
	if ((ADR((*memgetb)((*memgetb)(regs.PC+1)), (*memgetb)((*memgetb)(regs.PC+1))+1) & 0xFF00) !=	\
		((ADR((*memgetb)((*memgetb)(regs.PC+1)), (*memgetb)((*memgetb)(regs.PC+1))+1) + regs.Y) & 0xFF00)) \
		CONSUME(1)		
/*@}*/

/** @name Memory reading/writing macros - according to adressing mode */
/*@{*/
/** Immediate adressing mode */
#define IMM	\
	regs.PC+1
/** Zero page adressing mode */
#define ZP	\
	(*memgetb)(regs.PC+1)&0xFF
/** Zero page, X adressing mode */
#define ZPX	\
	((*memgetb)(regs.PC+1)+regs.X)&0xFF
/** Zero page, Y adressing mode */
#define ZPY	\
	((*memgetb)(regs.PC+1)+regs.Y)&0xFF
/** Absolute adressing mode */
#define ABS	\
	ADR((*memgetb)(regs.PC+1), (*memgetb)(regs.PC+2))
/** Absolute, X adressing mode */
#define ABSX	\
	ADR((*memgetb)(regs.PC+1), (*memgetb)(regs.PC+2))+regs.X
/** Absolute, Y adressing mode */
#define ABSY	\
	ADR((*memgetb)(regs.PC+1), (*memgetb)(regs.PC+2))+regs.Y
/** Indirect, X adressing mode */
#define INDX	\
	ADR((*memgetb)(((*memgetb)(regs.PC+1)+regs.X) & 0xFF), (*memgetb)(((*memgetb)(regs.PC+1)+regs.X+1) & 0xFF))
/** Indirect, Y adressing mode */
#define INDY	\
	ADR((*memgetb)((*memgetb)(regs.PC+1)), (*memgetb)(((*memgetb)(regs.PC+1)+1)) & 0xFF) + regs.Y
/** Relative adressing mode */
#define REL	\
	regs.PC+1
/** Used only for jmp, should be absolute */
#define JMPADR	\
	ADR((*memgetb)(regs.PC+1), (*memgetb)(regs.PC+2))
/*@}*/

/**
 * @name Instruction defs
 * Format: INS(adressing mode macro, lenght of instruction)
 */

/*@{*/
/** ADC: [A] + [M] + [C] -> [A], sets NVZC */
#define ADC(MEMREAD, STEP)   \
	if (regs.P & FD) {	\
		tmp=HEX2DEC(regs.A)+HEX2DEC((*memgetb)(MEMREAD))+GETFLAG(FC); \
		if( tmp > 99 ) {	\
			tmp=DEC2HEX(tmp);	\
			SETFLAG(FC);	\
			if ((tmp & 0x40)==(regs.A & 0x40)) SETFLAG(FV);	\
		}	\
		else {	\
			tmp=DEC2HEX(tmp);	\
			if (!(tmp & 0x40) && (regs.A & 0x40)) SETFLAG(FV); \
		}	\
		regs.A=tmp; \
	}	\
	else {	\
		tmp=regs.A+(*memgetb)(MEMREAD)+GETFLAG(FC);	\
		if(tmp<regs.A) {	\
			SETFLAG(FC);	\
			if ((tmp & 0x40)==(regs.A & 0x40)) SETFLAG(FV);	\
		}	\
		else {	\
			if (!(tmp & 0x40) && (regs.A & 0x40)) SETFLAG(FV); \
		}	\
		regs.A=tmp;	\
	}	\
	SETNZFLAGS(regs.A); \
	PCSTEP(STEP)

/** AND: [A] & [M] -> [A], sets NZ */
#define AND(MEMREAD, STEP)	\
	regs.A=regs.A & (*memgetb)(MEMREAD); \
	SETNZFLAGS(regs.A); \
	PCSTEP(STEP)

/** ASL: [C] <- [M] <- 0, left shift, sets NZC */
#define ASL(MEMREAD, STEP)	\
	tmp=(*memgetb)(MEMREAD);	\
	FLAG(FC, tmp & 0x80);	\
	tmp=(tmp << 1);	\
	(*memsetb)(MEMREAD, tmp);	\
	SETNZFLAGS(tmp);	\
	PCSTEP(STEP)

/** ASL: [C] <- [A] <- 0, left shift, sets NZ */
#define ASLA	\
	FLAG(FC, regs.A & 0x80);	\
	regs.A=(regs.A << 1);	\
	SETNZFLAGS(regs.A);	\
	PCSTEP(1)

/** BCC: relative jump, if [C]=0 */
#define BCC(MEMREAD, STEP)	\
	if (!GETFLAG(FC)) {	\
		CONSUME(3);	\
		tmp=(*memgetb)(MEMREAD);	\
		if (tmp>0x7F) {	\
			PCSTEP(STEP-0x100+tmp);	\
		}	\
		else {	\
			PCSTEP(STEP+tmp);	\
		}	\
	}	\
	else {	\
		PCSTEP(STEP);	\
		CONSUME(2);	\
	}

/** BCS: relative jump, if [C]=1 */
#define BCS(MEMREAD, STEP)	\
	if (GETFLAG(FC)) {	\
		CONSUME(3);	\
		tmp=(*memgetb)(MEMREAD);	\
		if (tmp>0x7F) {	\
			PCSTEP(STEP-0x100+tmp);	\
		}	\
		else {	\
			PCSTEP(STEP+tmp);	\
		}	\
	}	\
	else {	\
		PCSTEP(STEP);	\
		CONSUME(2);	\
	}

/** BEQ: relative jump, if [Z]=1 */
#define BEQ(MEMREAD, STEP)	\
	if (GETFLAG(FZ)) {	\
		CONSUME(3);	\
		tmp=(*memgetb)(MEMREAD);	\
		if (tmp>0x7F) {	\
			PCSTEP(STEP-0x100+tmp);	\
		}	\
		else {	\
			PCSTEP(STEP+tmp);	\
		}	\
	}	\
	else {	\
		PCSTEP(STEP);	\
		CONSUME(2);	\
	}

/** BNE: relative jump, if [Z]=0 */
#define BNE(MEMREAD, STEP)	\
	if (!GETFLAG(FZ)) {	\
		CONSUME(3);	\
		tmp=(*memgetb)(MEMREAD);	\
		if (tmp>0x7F) {	\
			PCSTEP(STEP-0x100+tmp);	\
		}	\
		else {	\
			PCSTEP(STEP+tmp);	\
		}	\
	}	\
	else {	\
		PCSTEP(STEP);	\
		CONSUME(2);	\
	}

/** BMI: relative jump, if [N]=1 */
#define BMI(MEMREAD, STEP)	\
	if (GETFLAG(FN)) {	\
		CONSUME(3);	\
		tmp=(*memgetb)(MEMREAD);	\
		if (tmp>0x7F) {	\
			PCSTEP(STEP-0x100+tmp);	\
		}	\
		else {	\
			PCSTEP(STEP+tmp);	\
		}	\
	}	\
	else {	\
		PCSTEP(STEP);	\
		CONSUME(2);	\
	}

/** BPL: relative jump, if [N]=0 */
#define BPL(MEMREAD, STEP)	\
	if (!GETFLAG(FN)) {	\
		CONSUME(3);	\
		tmp=(*memgetb)(MEMREAD);	\
		if (tmp>0x7F) {	\
			PCSTEP(STEP-0x100+tmp);	\
		}	\
		else {	\
			PCSTEP(STEP+tmp);	\
		}	\
	}	\
	else {	\
		PCSTEP(STEP);	\
		CONSUME(2);	\
	}

/** BVC: relative jump, if [V]=0 */
#define BVC(MEMREAD, STEP)	\
	if (!GETFLAG(FV)) {	\
		CONSUME(3);	\
		tmp=(*memgetb)(MEMREAD);	\
		if (tmp>0x7F) {	\
			PCSTEP(STEP-0x100+tmp);	\
		}	\
		else {	\
			PCSTEP(STEP+tmp);	\
		}	\
	}	\
	else {	\
		PCSTEP(STEP);	\
		CONSUME(2);	\
	}

/** BVS: relative jump, if [V]=1 */
#define BVS(MEMREAD, STEP)	\
	if (GETFLAG(FC)) {	\
		CONSUME(3);	\
		tmp=(*memgetb)(MEMREAD);	\
		if (tmp>0x7F) {	\
			PCSTEP(STEP-0x100+tmp);	\
		}	\
		else {	\
			PCSTEP(STEP+tmp);	\
		}	\
	}	\
	else {	\
		PCSTEP(STEP);	\
		CONSUME(2);	\
	}

/** BIT: [A] and [M], [M7] -> [N], [M6] -> [V], result sets [Z] */
#define BIT(MEMREAD, STEP)	\
	tmp=(*memgetb)(MEMREAD);	\
	FLAG(FN, tmp & 0x80);	\
	FLAG(FV, tmp & 0x40);	\
	tmp=tmp & regs.A;	\
	FLAG(FZ, ~tmp);	\
	PCSTEP(STEP)

/** BRK: Sw interrupt, 1 -> [B], [PC] -> Stack, [P] -> Stack, 1 -> [I]
 * 0 -> [D] (According to Western Design Center W65C02), IRQVKT -> PC
 */
#define BRK	\
	PCSTEP(2);	\
	SETFLAG(FB);	\
	STACKADD(HI(regs.PC));	\
	STACKADD(LO(regs.PC));	\
	STACKADD(regs.P);	\
	SETFLAG(FI);	\
	CLFLAG(FD);	\
	M2PC(IRQVKT)

/** CLC: 0 -> [C] */
#define CLC \
	CLFLAG(FC);	\
	PCSTEP(1)

/** CLD: 0 -> [D] */
#define CLD	\
	CLFLAG(FD);	\
	PCSTEP(1)

/** CLI: 0 -> [I] */
#define CLI	\
	CLFLAG(FI);	\
	PCSTEP(1)

/** CLV: 0 -> [V] */
#define CLV	\
	CLFLAG(FV);	\
	PCSTEP(1)

/** CMP: [A] - [M], result sets NZC */
#define CMP(MEMREAD, STEP)	\
	tmp=(*memgetb)(MEMREAD);	\
	if (regs.A<tmp) SETFLAG(FC);	\
	SETNZFLAGS(regs.A-tmp);	\
	PCSTEP(STEP)

/** CPX: [X] - [M], result sets NZC */
#define CPX(MEMREAD, STEP)	\
	tmp=(*memgetb)(MEMREAD);	\
	if (regs.X<tmp) SETFLAG(FC);	\
	SETNZFLAGS(regs.X-tmp);	\
	PCSTEP(STEP)

/** CPY: [Y] - [M], result sets NZC */
#define CPY(MEMREAD, STEP)	\
	tmp=(*memgetb)(MEMREAD);	\
	if (regs.Y<tmp) SETFLAG(FC);	\
	SETNZFLAGS(regs.Y-tmp);	\
	PCSTEP(STEP)

/** DEC: [M] - 1 -> [M], sets NZ */
#define DEC(MEMREAD, STEP)	\
	tmp=(*memgetb)(MEMREAD);	\
	tmp--;	\
	SETNZFLAGS(tmp);	\
	(*memsetb)(MEMREAD, tmp);	\
	PCSTEP(STEP)

/** DEX: [X] -1 -> [X], sets NZ */
#define DEX	\
	regs.X--;	\
	SETNZFLAGS(regs.X);	\
	PCSTEP(1)

/** DEY: [Y] - 1 -> [Y], sets NZ */
#define DEY	\
	regs.Y--;	\
	SETNZFLAGS(regs.Y);	\
	PCSTEP(1)

/** EOR: [A] xor [M] -> [A], sets NZ */
#define EOR(MEMREAD, STEP)	\
	regs.A=regs.A ^ (*memgetb)(MEMREAD);	\
	SETNZFLAGS(regs.A);	\
	PCSTEP(STEP)

/** INC: [M] + 1 -> [M], sets NZ */
#define INC(MEMREAD, STEP)	\
	tmp=(*memgetb)(MEMREAD);	\
	tmp++;	\
	SETNZFLAGS(tmp);	\
	(*memsetb)(MEMREAD, tmp);	\
	PCSTEP(STEP)

/** INX: [X] + 1 -> [X], sets NZ */
#define INX	\
	regs.X++;	\
	SETNZFLAGS(regs.X);	\
	PCSTEP(1)

/** INY: [Y] + 1 -> [Y], sets NZ */
#define INY	\
	regs.Y++;	\
	SETNZFLAGS(regs.Y);	\
	PCSTEP(1)

/** JMP: [M] -> [PC]
 * JMP(..) only for ABSOLUTE adressing! Explanation follows 
 */
#define JMP(MEMREAD)	\
				regs.PC=MEMREAD

/** Indirect jump must never use adress beginning on 0xXXFF.
 * if so, jump will be done to adress stored in 0xXXFF and 0xXX00 ! (no page overflow) 
 */
#define JMPIND	\
	regs.PC=ADR(	\
		(*memgetb)(ADR((*memgetb)(regs.PC+1), (*memgetb)(regs.PC+2))),	\
		(*memgetb)(ADR(((*memgetb)(regs.PC+1)+1) & 0xFF, (*memgetb)(regs.PC+2))))

/** JSR: [PC] + 2 - 1 -> Stack, [M] -> PC
 * JSR puts adress of next instruction-1 on the stack (rts adds +1) 
 */
#define JSR(MEMREAD)	\
	tmpjsr=MEMREAD;	\
	PCSTEP(2);	\
	STACKADD(HI(regs.PC));	\
	STACKADD(LO(regs.PC));	\
	regs.PC=tmpjsr

/** LDA: [M] -> [A], sets NZ */
#define LDA(MEMREAD, STEP)	\
	regs.A=(*memgetb)(MEMREAD);	\
	SETNZFLAGS(regs.A);	\
	PCSTEP(STEP)

/** LDX: [M] -> [X], sets NZ */
#define LDX(MEMREAD, STEP)	\
	regs.X=(*memgetb)(MEMREAD);	\
	SETNZFLAGS(regs.X);	\
	PCSTEP(STEP)

/** LDY: [M] -> [Y], sets NZ */
#define LDY(MEMREAD, STEP)	\
	regs.Y=(*memgetb)(MEMREAD);	\
	SETNZFLAGS(regs.Y);	\
	PCSTEP(STEP)

/** LSR: 0 -> [M] -> [C], right shift, sets: NZC */
#define LSR(MEMREAD, STEP)	\
	tmp=(*memgetb)(MEMREAD);	\
	FLAG(FC, tmp & 0x01);	\
	tmp=(tmp >> 1);	\
	(*memsetb)(MEMREAD, tmp);	\
	SETNZFLAGS(tmp);	\
	PCSTEP(STEP)

/** LSR: 0 -> [A] -> [C], right shift, sets: NZC */
#define LSRA	\
	FLAG(FC, regs.A & 0x01);	\
	regs.A=(regs.A >> 1);	\
	SETNZFLAGS(regs.A);	\
	PCSTEP(1)

/** NOP: Reserved */
#define NOP	\
	PCSTEP(1)

/** ORA: [A] or [M] -> [A], sets NZ */
#define ORA(MEMREAD, STEP)	\
	regs.A=regs.A | (*memgetb)(MEMREAD);	\
	SETNZFLAGS(regs.A);	\
	PCSTEP(STEP)

/** PHA: A -> Stack */
#define PHA	\
	STACKADD(regs.A);	\
	PCSTEP(1)

/** PHP: P -> Stack */
#define PHP	\
	STACKADD(regs.P);	\
	PCSTEP(1)

/** PLA: Stack -> A, sets NZ */
#define PLA	\
	STACKREMOVE(regs.A);	\
	SETNZFLAGS(regs.A);	\
	PCSTEP(1)

/** PLP: Stack -> P, sets NVBDIZC */
#define PLP	\
	STACKREMOVE(regs.P);	\
	regs.P=regs.P | 0x20;	\
	PCSTEP(1)

/** ROL: [C_new] <- [M] <- [C_old], left rotation, sets NZC */
#define ROL(MEMREAD, STEP)	\
	tmp=(*memgetb)(MEMREAD);	\
	tmpc=GETFLAG(FC);	\
	FLAG(FC, tmp & 0x80);	\
	tmp=(tmp << 1) | ((tmpc) ? 0x01 : 0x00);	\
	(*memsetb)(MEMREAD, tmp);	\
	PCSTEP(STEP)

/** ROL: [C_new] <- [A] <- [C_old], left rotation, sets NZC */
#define ROLA	\
	tmpc=GETFLAG(FC);	\
	FLAG(FC, regs.A & 0x80);	\
	regs.A=(regs.A << 1) | ((tmpc) ? 0x01 : 0x00);	\
	PCSTEP(1)

/** ROR: [C_old] -> [M] -> [C_new], right rotation, sets NZC */
#define ROR(MEMREAD, STEP)	\
	tmp=(*memgetb)(MEMREAD);	\
	tmpc=GETFLAG(FC);	\
	FLAG(FC, tmp & 0x01);	\
	tmp=(tmp >> 1) | ((tmpc) ? 0x80 : 0x00);	\
	(*memsetb)(MEMREAD, tmp);	\
	PCSTEP(STEP)

/** ROR A: [C_old] -> [A] -> [C_new], right rotation, sets NZC */
#define RORA	\
	tmpc=GETFLAG(FC);	\
	FLAG(FC, regs.A & 0x01);	\
	regs.A=(regs.A >> 1) | ((tmpc) ? 0x80 : 0x00);	\
	PCSTEP(1)

/** RTI: Stack -> [P], Stack -> [PC], return from interrupt, sets NVBDIZC */
#define RTI	\
	STACKREMOVE(regs.P);	\
	STACKREMOVE(tmp);	\
	STACKREMOVE(regs.PC);	\
	regs.PC=(regs.PC << 8) + tmp

/** RTS: Stack + 1 -> [PC], return from JSR subroutine  (for +1 see JSR) */
#define RTS	\
	STACKREMOVE(tmp);	\
	STACKREMOVE(regs.PC);	\
	regs.PC=(regs.PC << 8) + tmp + 1

/** SBC: [A] - [M] - non[C] -> [A], sets NVZ(non C) */
#define SBC(MEMREAD, STEP)	\
	if (regs.P & FD) {	\
		tmp=HEX2DEC(regs.A)-HEX2DEC((*memgetb)(MEMREAD))- (1-GETFLAG(FC)); \
		if(tmp<=99) {	\
			tmp=DEC2HEX(tmp);	\
			SETFLAG(FC);	\
			if ((tmp & 0x40) && !(regs.A & 0x40)) SETFLAG(FV);	\
		}	\
		else {	\
			tmp=tmp-156;	\
			tmp=DEC2HEX(tmp);	\
			if ((tmp & 0x40)==(regs.A & 0x40)) SETFLAG(FV); \
		}	\
		regs.A=tmp; \
	}	\
	else {	\
		tmp=regs.A-(*memgetb)(MEMREAD)-(1-GETFLAG(FC));	\
		if(tmp<=regs.A) {	\
			SETFLAG(FC);	\
			if ((tmp & 0x40) && !(regs.A & 0x40)) SETFLAG(FV);	\
		}	\
		else {	\
			if ((tmp & 0x40)==(regs.A & 0x40)) SETFLAG(FV); \
		}	\
		regs.A=tmp;	\
	}	\
	SETNZFLAGS(regs.A); \
	PCSTEP(STEP)

/** SEC: 1 -> [C] */
#define SEC	\
	SETFLAG(FC);	\
	PCSTEP(1)

/** SED: 1 -> [D] */
#define SED	\
	SETFLAG(FD);	\
	PCSTEP(1)

/** SEI: 1 -> [I] */
#define SEI	\
	SETFLAG(FI);	\
	PCSTEP(1)

/** STA: [A] -> [M] */
#define STA(MEMREAD, STEP)	\
	(*memsetb)(MEMREAD, regs.A);	\
	PCSTEP(STEP)

/** STX: [X] -> [M] */
#define STX(MEMREAD, STEP)	\
	(*memsetb)(MEMREAD, regs.X);	\
	PCSTEP(STEP)

/** STY: [Y] -> [M] */
#define STY(MEMREAD, STEP)	\
	(*memsetb)(MEMREAD, regs.Y);	\
	PCSTEP(STEP)

/** TAX: [A] -> [X], sets NZ */
#define TAX	\
	regs.X=regs.A;	\
	SETNZFLAGS(regs.X);	\
	PCSTEP(1)

/** TXA: [X] -> [A], sets NZ */
#define TXA	\
	regs.A=regs.X;	\
	SETNZFLAGS(regs.A);	\
	PCSTEP(1)

/** TAY: [A] -> [Y], sets NZ */
#define TAY	\
	regs.Y=regs.A;	\
	SETNZFLAGS(regs.Y);	\
	PCSTEP(1)

/** TYA: [Y] -> [A], sets NZ */
#define TYA	\
	regs.A=regs.Y;	\
	SETNZFLAGS(regs.A);	\
	PCSTEP(1)

/** TSX: [S] -> [X], sets NZ */
#define TSX	\
	regs.X=regs.S;	\
	SETNZFLAGS(regs.X);	\
	PCSTEP(1)

/** TXS: [X] -> [S] */
#define TXS	\
	regs.S=regs.X;	\
	PCSTEP(1)
/*@}*/

/** Variable containing all registers */
Regs regs;

/**
 * @brief Executes one instruction on adress stored in @a regs.PC
 * 
 * Executes one instruction on adress stored in @a regs.PC;
 * increments (or changes in case of jump) program counter
 * @return Number of cycles consumed
 * @todo Add undocumented instructions
 */
UINT InstrStep()
{
	unsigned char tmp, tmpc;
	UINT tmpjsr;
	UINT oldpc = regs.PC;
	UINT cycles = 0;
	switch((*memgetb)(regs.PC)) {
	case 0:
		BRK;
		CONSUME(7);
		break;
	case 1:
		ORA(INDX, 2);
		CONSUME(6);
		break;
	case 5:
		ORA(ZP, 2);
		CONSUME(3);
		break;
	case 6:
		ASL(ZP, 2);
		CONSUME(5);
		break;
	case 8:
		PHP;
		CONSUME(3);
		break;
	case 9:
		ORA(IMM, 2);
		CONSUME(2);
		break;
	case 10:
		ASLA;
		CONSUME(2);
		break;
	case 13:
		ORA(ABS, 3);
		CONSUME(4);
		break;
	case 14:
		ASL(ABS, 3);
		CONSUME(6);
		break;
	case 16:
		BPL(IMM, 2);	//relative; cycles consumed in macro
		CONSREL;
		break;
	case 17:
		CONSINDY;
		ORA(INDY, 2);
		CONSUME(4);
		break;
	case 21:
		ORA(ZPX, 2);
		CONSUME(4);
		break;
	case 22:
		ASL(ZPX, 2);
		CONSUME(6);
		break;
	case 24:
		CLC;
		CONSUME(2);
		break;
	case 25:
		CONSABSY;
		ORA(ABSY, 3);
		CONSUME(4);
		break;
	case 29:
		CONSABSX;
		ORA(ABSX, 3);
		CONSUME(4);
		break;
	case 30:
		ASL(ABSX, 3);
		CONSUME(7);
		break;
	case 32:
		JSR(JMPADR);
		CONSUME(6);
		break;
	case 33:
		AND(INDX, 2);
		CONSUME(6);
		break;
	case 36:
		BIT(ZP, 2);
		CONSUME(3);
		break;
	case 37:
		AND(ZP, 2);
		CONSUME(3);
		break;
	case 38:
		ROL(ZP, 2);
		CONSUME(5);
		break;
	case 40:
		PLP;
		CONSUME(4);
		break;
	case 41:
		AND(IMM, 2);
		CONSUME(2);
		break;
	case 42:
		ROLA;
		CONSUME(2);
		break;
	case 44:
		BIT(ABS, 3);
		CONSUME(4);
		break;
	case 45:
		AND(ABS, 3);
		CONSUME(4);
		break;
	case 46:
		ROL(ABS, 3);
		CONSUME(6);
		break;
	case 48:
		BMI(IMM, 2);	//relative; cycles consumed in macro
		CONSREL;
		break;
	case 49:
		CONSINDY;
		AND(INDY, 2);
		CONSUME(5);
		break;
	case 53:
		AND(ZPX, 2);
		CONSUME(4);
		break;
	case 54:
		ROL(ZPX, 2);
		CONSUME(6);
		break;
	case 56:
		SEC;
		CONSUME(2);
		break;
	case 57:
		CONSABSY;
		AND(ABSY, 3);
		CONSUME(4);
		break;
	case 61:
		CONSABSX;
		AND(ABSX, 3);
		CONSUME(4);
		break;
	case 62:
		ROL(ABSX, 3);
		CONSUME(7);
		break;
	case 64:
		RTI;
		CONSUME(6);
		break;
	case 65:
		EOR(INDX, 2);
		CONSUME(6);
		break;
	case 69:
		EOR(ZP, 2);
		CONSUME(3);
		break;
	case 70:
		LSR(ZP, 2);
		CONSUME(5);
		break;
	case 72:
		PHA;
		CONSUME(3);
		break;
	case 73:
		EOR(IMM, 2);
		CONSUME(2);
		break;
	case 74:
		LSRA;
		CONSUME(2);
		break;
	case 76:
		JMP(JMPADR);
		CONSUME(3);
		break;
	case 77:
		EOR(ABS, 3);
		CONSUME(4);
		break;
	case 78:
		LSR(ABS, 3);
		CONSUME(6);
		break;
	case 80:
		BVC(IMM, 2);	//relative; cycles consumed in macro
		CONSREL;
		break;
	case 81:
		CONSINDY;
		EOR(INDY, 2);
		CONSUME(5);
		break;
	case 85:
		EOR(ZPX, 2);
		CONSUME(4);
		break;
	case 86:
		LSR(ZPX, 2);
		CONSUME(6);
		break;
	case 88:
		CLI;
		CONSUME(2);
		break;
	case 89:
		CONSABSY;
		EOR(ABSY, 3);
		CONSUME(4);
		break;
	case 93:
		CONSABSX;
		EOR(ABSX, 3);
		CONSUME(4);
		break;
	case 94:
		LSR(ABSX, 3);
		CONSUME(7);
		break;
	case 96:
		RTS;
		CONSUME(6);
		break;
	case 97:
		ADC(INDX, 2);
		CONSUME(6);
		break;
	case 101:
		ADC(ZP, 2);
		CONSUME(3);
		break;
	case 102:
		ROR(ZP, 2);
		CONSUME(5);
		break;
	case 104:
		PLA;
		CONSUME(4);
		break;
	case 105:
		ADC(IMM, 2);
		CONSUME(2);
		break;
	case 106:
		RORA;
		CONSUME(2);
		break;
	case 108:
		JMPIND;
		CONSUME(5);
		break;
	case 109:
		ADC(ABS, 3);
		CONSUME(4);
		break;
	case 110:
		ROR(ABS, 3);
		CONSUME(2);
		break;
	case 112:
		BVS(IMM, 2);	//relative; cycles consumed in macro
		CONSREL;
		break;
	case 113:
		CONSINDY;
		ADC(INDY, 2);
		CONSUME(4);
		break;
	case 117:
		ADC(ZPX, 2);
		CONSUME(4);
		break;
	case 118:
		ROR(ZPX, 2);
		CONSUME(6);
		break;
	case 120:
		SEI;
		CONSUME(2);
		break;
	case 121:
		CONSABSY;
		ADC(ABSY, 3);
		CONSUME(4);
		break;
	case 125:
		CONSABSX;
		ADC(ABSX, 3);
		CONSUME(4);
		break;
	case 126:
		ROR(ABSX, 3);
		CONSUME(7);
		break;
	case 129:
		STA(INDX, 2);
		CONSUME(6);
		break;
	case 132:
		STY(ZP, 2);
		CONSUME(3);
		break;
	case 133:
		STA(ZP, 2);
		CONSUME(3);
		break;
	case 134:
		STX(ZP, 2);
		CONSUME(3);
		break;
	case 136:
		DEY;
		CONSUME(2);
		break;
	case 138:
		TXA;
		CONSUME(2);
		break;
	case 140:
		STY(ABS, 3);
		CONSUME(4);
		break;
	case 141:
		STA(ABS, 3);
		CONSUME(4);
		break;
	case 142:
		STX(ABS, 3);
		CONSUME(4);
		break;
	case 144:
		BCC(IMM, 2);	//relative; cycles consumed in macro
		CONSREL;
		break;
	case 145:
		STA(INDY, 2);
		CONSUME(6);
		break;
	case 148:
		STY(ZPX, 2);
		CONSUME(4);
		break;
	case 149:
		STA(ZPX, 2);
		CONSUME(4);
		break;
	case 150:
		STX(ZPX, 2);
		CONSUME(4);
		break;
	case 152:
		TYA;
		CONSUME(2);
		break;
	case 153:
		STA(ABSY, 3);
		CONSUME(5);
		break;
	case 154:
		TXS;
		CONSUME(2);
		break;
	case 157:
		STA(ABSX, 3);
		CONSUME(5);
		break;
	case 160:
		LDY(IMM, 2);
		CONSUME(2);
		break;
	case 161:
		LDA(INDX, 2);
		CONSUME(6);
		break;
	case 162:
		LDX(IMM, 2);
		CONSUME(2);
		break;
	case 164:
		LDY(ZP, 2);
		CONSUME(3);
		break;
	case 165:
		LDA(ZP, 2);
		CONSUME(3);
		break;
	case 166:
		LDX(ZP, 2);
		CONSUME(3);
		break;
	case 168:
		TAY;
		CONSUME(2);
		break;
	case 169:
		LDA(IMM, 2);
		CONSUME(2);
		break;
	case 170:
		TAX;
		CONSUME(2);
		break;
	case 172:
		LDY(ABS, 3);
		CONSUME(4);
		break;
	case 173:
		LDA(ABS, 3);
		CONSUME(4);
		break;
	case 174:
		LDX(ABS, 3);
		CONSUME(4);
		break;
	case 175:
		BCS(IMM, 2);	//relative; cycles consumed in macro
		CONSREL;
		break;
	case 177:
		CONSINDY;
		LDA(INDY, 2);
		CONSUME(5);
		break;
	case 180:
		LDY(ZPX, 2);
		CONSUME(4);
		break;
	case 181:
		LDA(ZPX, 2);
		CONSUME(4);
		break;
	case 182:
		LDX(ZPY, 2);
		CONSUME(4);
		break;
	case 184:
		CLV;
		CONSUME(2);
		break;
	case 185:
		CONSABSY;
		LDA(ABSY, 3);
		CONSUME(4);
		break;
	case 186:
		TSX;
		CONSUME(2);
		break;
	case 188:
		CONSABSY;
		LDY(ABSX, 3);
		CONSUME(4);
		break;
	case 189:
		CONSABSX;
		LDA(ABSX, 3);
		CONSUME(4);
		break;
	case 190:
		CONSABSY;
		LDX(ABSY, 3);
		CONSUME(4);
		break;
	case 192:
		CPY(IMM, 2);
		CONSUME(2);
		break;
	case 193:
		CMP(INDX, 2);
		CONSUME(6);
		break;
	case 196:
		CPY(ZP, 2);
		CONSUME(3);
		break;
	case 197:
		CMP(ZP, 2);
		CONSUME(3);
		break;
	case 198:
		DEC(ZP, 2);
		CONSUME(5);
		break;
	case 200:
		INY;
		CONSUME(2);
		break;
	case 201:
		CMP(IMM, 2);
		CONSUME(2);
		break;
	case 202:
		DEX;
		CONSUME(2);
		break;
	case 204:
		CPY(ABS, 3);
		CONSUME(4);
		break;
	case 205:
		CMP(ABS, 3);
		CONSUME(4);
		break;
	case 206:
		DEC(ABS, 3);
		CONSUME(6);
		break;
	case 208:
		BNE(IMM, 2);	//relative; cycles consumed in macro
		CONSREL;
		break;
	case 209:
		CONSINDY;
		CMP(INDY, 2);
		CONSUME(5);
		break;
	case 213:
		CMP(ZPX, 2);
		CONSUME(4);
		break;
	case 214:
		DEC(ZPX, 2);
		CONSUME(6);
		break;
	case 216:
		CLD;
		CONSUME(2);
		break;
	case 217:
		CONSABSY;
		CMP(ABSY, 3);
		CONSUME(4);
		break;
	case 221:
		CONSABSX;
		CMP(ABSX, 3);
		CONSUME(4);
		break;
	case 222:
		DEC(ABSX, 3);
		CONSUME(7);
		break;
	case 224:
		CPX(IMM, 2);
		CONSUME(2);
		break;
	case 225:
		SBC(INDX, 2);
		CONSUME(6);
		break;
	case 228:
		CPX(ZP, 2);
		CONSUME(3);
		break;
	case 229:
		SBC(ZP, 2);
		CONSUME(3);
		break;
	case 230:
		INC(ZP, 2);
		CONSUME(5);
		break;
	case 232:
		INX;
		CONSUME(2);
		break;
	case 233:
		SBC(IMM, 2);
		CONSUME(2);
		break;
	case 234:
		NOP;
		CONSUME(2);
		break;
	case 236:
		CPX(ABS, 3);
		CONSUME(4);
		break;
	case 237:
		SBC(ABS, 3);
		CONSUME(4);
		break;
	case 238:
		INC(ABS, 3);
		CONSUME(6);
		break;
	case 240:
		BEQ(IMM, 2);	//relative; cycles consumed in macro
		CONSREL;
		break;
	case 241:
		CONSINDY;
		SBC(INDY, 2);
		CONSUME(5);
		break;
	case 245:
		SBC(ZPX, 2);
		CONSUME(4);
		break;
	case 248:
		SED;
		CONSUME(2);
		break;
	case 249:
		CONSABSY;
		SBC(ABSY, 3);
		CONSUME(4);
		break;
	case 253:
		CONSABSX;
		SBC(ABSX, 3);
		CONSUME(4);
		break;
	case 254:
		INC(ABSX, 3);
		CONSUME(7);
		break;
	default:
		PCSTEP(1);
		CONSUME(2);		//like NOP
	}
	regs.PC = regs.PC & 0xFFFF;
	return cycles;
}

Regs* GetRegs()
{
	return &regs;
}

void SetRegs(Regs r)
{
	regs = r;
}

/**
 * @brief Reset interrupt
 * 
 * Zeroes A, X, Y registers, resets stack pointer S, 
 * initializes P register (zeroes all flags except I, which is set to 1)
 * and jumps to the @a RESETVKT, reset initialization routine.
 */

void Reset()
{
	regs.A=0;
	regs.X=0;
	regs.Y=0;
	regs.P=0x20 | FI;   // 6th bit is always set to 1; set I to 1
	regs.S=0x1FF;		// initialise stack pointer
	M2PC(RESETVKT);
}

/**
 * @brief IRQ interrupt
 * 
 * Adds actual position and the P register (flags) to the top of the stack,
 * sets the I (interrupt) flag to 1
 * and jumps to the @a IRQVKT, reset initialization routine.
 */
void Irq()
{
	if (regs.P & FI != 0) {		// Interrupt is not allowed
		return;
	}
	STACKADD(HI(regs.PC));
	STACKADD(LO(regs.PC));
	STACKADD(regs.P);
	SETFLAG(FI);
	M2PC(IRQVKT);
}

/**
 * @brief NMI interrupt
 * 
 * Adds actual position and the P register (flags) to the top of the stack,
 * sets the I (interrupt) flag to 1
 * and jumps to the @a NMIVKT, reset initialization routine.
 */
void Nmi()
{
	STACKADD(HI(regs.PC));
	STACKADD(LO(regs.PC));
	STACKADD(regs.P);
	SETFLAG(FI);
	M2PC(NMIVKT);
}