/**
 * @file cpuuidll.c
 * Menues and toolbar buttons for controlling cpu
 */

#include "../dllkit/dllkit.h"
#include "../usrmsgs.h"
#include "resource.h"
#include <commctrl.h>
#include <stdio.h>
#include <tchar.h>

/** @brief Handle to instance of this module */
HANDLE hThisModule;
/** @brief Contains list of messages to be forwarded to this plugin */
MSGLIST Msgs;
/** @brief Local copy of a structure passed to this plugin containging
 * various info on the main emu6502 window and application */
INITDLLSTRUCT info;

/** @brief Handle to the main menu bar */ 
HMENU hMenu;
/** @brief Handle to the CPU menu */
HMENU hCpuMenu;

BOOL Running;

/* ----- begin forward declarations */
/** @brief enables/disables accordant toolbar buttons and menu entries */
void SetStatusRunning();
/** @brief enables/disables accordant toolbar buttons and menu entries */
void SetStatusStopped();
/* ----- end forward declarations */

/** @brief Entry point of this dll */
BOOL APIENTRY DllMain(HANDLE hModule, DWORD  ul_reason_for_call, 
					  LPVOID lpReserved)
{
	hThisModule = hModule;
	switch (ul_reason_for_call) {
	case DLL_PROCESS_ATTACH:
		CreateMsgList(&Msgs, 3);
		break;
	case DLL_PROCESS_DETACH:
		FreeMsgList(Msgs);
		break;
	}	
    return TRUE;
}

DLLEXPORT void __cdecl InitDll(INITDLLSTRUCT* ids, INITDLLINFO* idi)
{
	TBADDBITMAP tbab;
	TBBUTTON tbt[7];
	int index;

	// Fill the msglist with messages we want handle
	AddMessage(Msgs, WM_COMMAND);
	AddMessage(Msgs, CPURUNNING);
	AddMessage(Msgs, CPUSTOPPED);

	// Copy informations passed to this dll
	info.CSMemory = ids->CSMemory;
	info.hInst = ids->hInst;
	info.hWnd = ids->hWnd;
	info.memory = ids->memory;
	info.memgetb = ids->memgetb;
	info.memsetb = ids->memsetb;
	info.optional = ids->optional;
	info.hTb = ids->hTb;

	// Append local menu to the main menu bar
	hMenu = GetMenu(info.hWnd);
	hCpuMenu = LoadMenu(hThisModule, MAKEINTRESOURCE(IDR_MENU));
	AppendMenu(hMenu, MF_STRING | MF_POPUP, (UINT) hCpuMenu, "&CPU");
	DrawMenuBar(info.hWnd);

	// Add buttons to the main toolbar
	tbab.hInst = hThisModule;
	tbab.nID = IDR_TOOLBAR;
	index = SendMessage(info.hTb, TB_ADDBITMAP, (WPARAM) 5, 
        (WPARAM) &tbab); 
	ZeroMemory(&tbt, sizeof(TBBUTTON) * 5);
	tbt[0].iBitmap = index;
	tbt[0].idCommand = ID_RESUME;
	tbt[0].fsState = TBSTATE_ENABLED;
	tbt[0].fsStyle = TBSTYLE_BUTTON | TBSTYLE_AUTOSIZE;
	tbt[1].iBitmap = index + 1;
	tbt[1].idCommand = ID_PAUSE;
	tbt[1].fsState = TBSTATE_ENABLED;
	tbt[1].fsStyle = TBSTYLE_BUTTON | TBSTYLE_AUTOSIZE;
	tbt[2].iBitmap = index + 2;
	tbt[2].idCommand = ID_NMI;
	tbt[2].fsState = TBSTATE_ENABLED;
	tbt[2].fsStyle = TBSTYLE_BUTTON | TBSTYLE_AUTOSIZE;
	tbt[3].iBitmap = index + 3;
	tbt[3].idCommand = ID_IRQ;
	tbt[3].fsState = TBSTATE_ENABLED;
	tbt[3].fsStyle = TBSTYLE_BUTTON | TBSTYLE_AUTOSIZE;
	tbt[4].iBitmap = index + 4;
	tbt[4].idCommand = ID_RESET;
	tbt[4].fsState = TBSTATE_ENABLED;
	tbt[4].fsStyle = TBSTYLE_BUTTON | TBSTYLE_AUTOSIZE;

    SendMessage(info.hTb, TB_ADDBUTTONS, (WPARAM) 5, 
        (LPARAM) &tbt); 

	SendMessage(info.hTb, TB_ENABLEBUTTON, 
		(WPARAM) ID_PAUSE, (LPARAM) MAKELONG(FALSE, 0)); 
	SendMessage(info.hTb, TB_ENABLEBUTTON, 
		(WPARAM) ID_RESET, (LPARAM) MAKELONG(FALSE, 0)); 
	SendMessage(info.hTb, TB_ENABLEBUTTON, 
		(WPARAM) ID_IRQ, (LPARAM) MAKELONG(FALSE, 0)); 
	SendMessage(info.hTb, TB_ENABLEBUTTON, 
		(WPARAM) ID_NMI, (LPARAM) MAKELONG(FALSE, 0)); 

    SendMessage(info.hTb, TB_AUTOSIZE, 0, 0); 

	Running = FALSE;

	idi->iolo = 0;
	idi->iohi = 0;
	idi->msglist = Msgs;
}

DLLEXPORT void __cdecl OnMessage(MESSAGEDLLSTRUCT* mds)
{
	int wmId, wmEvent;

	switch (mds->message) {
	case WM_COMMAND:
		wmId = LOWORD(mds->wParam); 
		wmEvent = HIWORD(mds->wParam); 
		switch (wmId) {
		case ID_RESUME:
			SetStatusRunning();
			SendMessage(info.hWnd, STARTCPU, 0, 0);
			break;		
		case ID_PAUSE:
			SetStatusStopped();
			SendMessage(info.hWnd, PAUSECPU, 0, 0);
			break;
		case ID_RESET:
			(*info.memsetb)(RESETID, 255);
			SendMessage(info.hWnd, RESETMSG, 0, 0);
			break;
		case ID_IRQ:
			(*info.memsetb)(IRQID, 255);
			SendMessage(info.hWnd, IRQMSG, 0, 0);
			break;
		case ID_NMI:
			(*info.memsetb)(NMIID, 255);
			SendMessage(info.hWnd, NMIMSG, 0, 0);
			break;
		case ID_ABOUT:
			MessageBox(info.hWnd, "Cpu UI\nAdds the CPU menu and buttons which enable sending commands to cpu from GUI\n(c)2002 Petr Cermak", "About", MB_OK);
			break;
		}
		break;
	case CPURUNNING:
		if (!Running) {
			SetStatusRunning();
		}
		break;
	case CPUSTOPPED:
		if (Running) {
			SetStatusStopped();
		}
		break;
	}

}

void SetStatusRunning()
{
	Running = TRUE;
	EnableMenuItem(hCpuMenu, ID_RESUME, MF_GRAYED | MF_BYCOMMAND);
	EnableMenuItem(hCpuMenu, ID_PAUSE, MF_ENABLED | MF_BYCOMMAND);
	EnableMenuItem(hCpuMenu, ID_RESET, MF_ENABLED | MF_BYCOMMAND);
	EnableMenuItem(hCpuMenu, ID_IRQ, MF_ENABLED | MF_BYCOMMAND);
	EnableMenuItem(hCpuMenu, ID_NMI, MF_ENABLED | MF_BYCOMMAND);
	SendMessage(info.hTb, TB_ENABLEBUTTON, 
		(WPARAM) ID_RESUME, (LPARAM) MAKELONG(FALSE, 0)); 
	SendMessage(info.hTb, TB_ENABLEBUTTON, 
		(WPARAM) ID_PAUSE, (LPARAM) MAKELONG(TRUE, 0)); 
	SendMessage(info.hTb, TB_ENABLEBUTTON, 
		(WPARAM) ID_RESET, (LPARAM) MAKELONG(TRUE, 0)); 
	SendMessage(info.hTb, TB_ENABLEBUTTON, 
		(WPARAM) ID_IRQ, (LPARAM) MAKELONG(TRUE, 0)); 
	SendMessage(info.hTb, TB_ENABLEBUTTON, 
		(WPARAM) ID_NMI, (LPARAM) MAKELONG(TRUE, 0)); 
}

void SetStatusStopped()
{
	Running = FALSE;
	EnableMenuItem(hCpuMenu, ID_RESUME, MF_ENABLED | MF_BYCOMMAND);
	EnableMenuItem(hCpuMenu, ID_PAUSE, MF_GRAYED | MF_BYCOMMAND);
	EnableMenuItem(hCpuMenu, ID_RESET, MF_GRAYED | MF_BYCOMMAND);
	EnableMenuItem(hCpuMenu, ID_IRQ, MF_GRAYED | MF_BYCOMMAND);
	EnableMenuItem(hCpuMenu, ID_NMI, MF_GRAYED | MF_BYCOMMAND);
	SendMessage(info.hTb, TB_ENABLEBUTTON, 
		(WPARAM) ID_RESUME, (LPARAM) MAKELONG(TRUE, 0)); 
	SendMessage(info.hTb, TB_ENABLEBUTTON, 
		(WPARAM) ID_PAUSE, (LPARAM) MAKELONG(FALSE, 0)); 
	SendMessage(info.hTb, TB_ENABLEBUTTON, 
		(WPARAM) ID_RESET, (LPARAM) MAKELONG(FALSE, 0)); 
	SendMessage(info.hTb, TB_ENABLEBUTTON, 
		(WPARAM) ID_IRQ, (LPARAM) MAKELONG(FALSE, 0)); 
	SendMessage(info.hTb, TB_ENABLEBUTTON, 
		(WPARAM) ID_NMI, (LPARAM) MAKELONG(FALSE, 0)); 
}

DLLEXPORT void __cdecl IOFunc(UINT adr, unsigned char* data, 
										  iomode mode)
{
}

