/**
 * @file dllkit.h
 * Header file of the 6502 dll developement kit.
 * Library for creating emu6502 plugins.
 */

#ifndef DLLKIT_H
#define DLLKIT_H

#include <windows.h>


#define DLLEXPORT __declspec(dllexport)

/** @brief Lower boundary of memory serving for device maping */
#define IOLO 0xD000
/** @brief Higher boundary of memory serving for device maping */
#define IOHI 0xD7FF

/** @brief Register in memory which serves to identify the source of last interrupt */
#define NMIID 0x020A
#define IRQID 0x020B
#define RESETID 0x020C
#define ISVBI 0x020D

/**
 * @brief List of messages
 *
 * Used for specifying messages to be catched or to be sent. 
 * Length of this list should be the number of messages
 * to be held within + 1, and the list should end with @a ENDOFMSGLIST
 */
typedef UINT* MSGLIST;

/** @brief Marks the end of the list of messages @a MSGLIST */
#define ENDOFMSGLIST 0xBFFF

/** @brief Structure containing various informations returned by @a InitDll */
typedef struct {
	MSGLIST msglist;	/**< @see MSGLIST */
	UINT iolo;	/**< Lower boundary of io memory reserved by this plugin */
	UINT iohi;  /**< Higher boundary of io memory reserved by this plugin */
} INITDLLINFO;

/**
 * @brief Initial informations passed to plugin when initializing.
 * 
 * Structure containing informations about the main application,
 * about 6502 memory and other.
 */
typedef struct {
	/** Pointer to shared memory. Use @a CSMemory when accessing it. */
	char* memory;
	/** Handle of the main window */
	HWND hWnd;
	/** Handle of the main window toolbar */
	HWND hTb;
	/** Instance of the emu6502 application */
	HINSTANCE hInst;
	/** Use when accesing shared @a memory */
	CRITICAL_SECTION CSMemory;
	/** Pointer to a function which gets a byte from given adress. 
	 * No need for synchronizing using @a CSMemory.*/
	unsigned char (*memgetb) (UINT adr);
	/** Pointer to a function which sets byte on given adress. 
	 * No need for synchronizing using @a CSMemory.*/
	void (*memsetb) (UINT adr, unsigned char byte);
	/** Area of allowed painting/refreshing */
	RECT rcPaint;
	/** Optional data passed by emu6502. */
	void* optional;
} INITDLLSTRUCT;

/**
 * @brief Message structure passed to plugins, when registered message occured.
 *
 * Structure containing informations about message sent to plugin.
 */
typedef struct {
	UINT message;
	WPARAM wParam;
	LPARAM lParam;
} MESSAGEDLLSTRUCT;


/**
 * @brief Allocates MsgList
 *
 * Allocates MsgList for holding MessageCount of messages. MsgList takes
 * (MessageCount + 1) * sizeof(UINT) bytes and first item is set to @a ENDOFMSGLIST.
 * @param MsgList A MSGLIST variable to be allocated.
 * @param MessageCount Max. count of messages to be held within the list.
 * @return TRUE if succesful, otherwise FALSE.
 */
BOOL CreateMsgList(MSGLIST* MsgList, int MessageCount);

/**
 * @brief Deallocates MsgList
 * Calls free on MsgList.
 * @param MsgList A MSGLIST variable to be deallocated.
 */
void FreeMsgList(MSGLIST MsgList);

/**
 * @brief Adds Message to the given MsgList
 *
 * Adds Message to the given MsgList. The message shouldn't be @a ENDOFMSGLIST.
 * @param MsgList List of messages to which the Message should be added.
 * @param Message Message which will be added to the MsgList
 */
void AddMessage(MSGLIST MsgList, UINT Message);


/** @brief Determines, if iofunc is called when reading or writing */
typedef enum { IOREAD, IOWRITE } iomode;

/**
 * @brief IO function mapped into some adress in memory
 *
 * Called directly when accesing adress to which is this function bound.
 * Always write this function. 
 * @param adr Adress being accessed
 * @param data Byte being read/written
 * @param mode see @a iomode
 */
DLLEXPORT void __cdecl IOFunc(UINT adr, unsigned char* data, iomode mode);


/**
 * @brief Inits dll data structures
 *
 * This function is called after dll is loaded. You have to implement
 * this function. Useful informations are passed in its parameter. 
 * Also create, fill with messages to be catched and return @a MSGLIST structure.
 * @param ids An @a INITDLLSTRUCT structure containig informations about 
 * calling emu6502 application.
 * @param idf An @a INITDLLINFO structure which should be filled in this function
 * <br><i>Note: if io memory boundaries to be reserved for this plugin are in conflict
 * with some other plugin, previous settings will be overwritten!</i>
 * @return Filled @a MSGLIST of messages to be catched and forwarded to this plugin
 * @see OnMessage(MESSAGEDLLSTRUCT* mds)
 * @see INITDLLSTRUCT
 * @see INITDLLINFO
 */
DLLEXPORT void __cdecl InitDll(INITDLLSTRUCT* ids, INITDLLINFO* idf);

/**
 * @brief Recieves registered messages from emu6502 application
 *
 * This function is called by emu6502 when some message event registered by
 * @a InitDll occures. Implement this function even when
 * you don't want to catch any messages.
 * @param mds A @a MESSAGEDLLSTRUCT structure contaning informations about
 * incoming message.
 * @see InitDll
 * @see MESSAGEDLLSTRUCT
 */
DLLEXPORT void __cdecl OnMessage(MESSAGEDLLSTRUCT* mds);

#endif	//#define DLLKIT_H
